// Copyright 2004 Adam Megacz, see the COPYING file for licensing [GPL] 
package org.ibex.js; 

import org.ibex.util.*; 
import java.io.*;
import java.util.*;

/** The minimum set of functionality required for objects which are manipulated by JavaScript */
public abstract class JS { 

    public static boolean checkAssertions = false;

    public static final JS METHOD = new JS() { };
    public final JS unclone() { return _unclone(); }
    public final JS jsclone() throws JSExn { return new Clone(this); }

    // FEATURE: JSEnumeration
    public Enumeration keys() throws JSExn { throw new JSExn("you can't enumerate the keys of this object (class=" + getClass().getName() +")"); }
    public JS get(JS key) throws JSExn { return null; }
    public void put(JS key, JS val) throws JSExn { throw new JSExn("" + key + " is read only (class=" + getClass().getName() +")"); }
        
    public final boolean hasTrap(JS key) { return getTrap(key) != null; }
    
    public JS callMethod(JS method, JS a0, JS a1, JS a2, JS[] rest, int nargs) throws JSExn {
        throw new JSExn("method not found (" + JS.debugToString(method) + ")");
    }    
    
    // FIXME: JSArgs objects, pointers into stack frame
    public JS call(JS a0, JS a1, JS a2, JS[] rest, int nargs) throws JSExn {
        throw new JSExn("you cannot call this object (class=" + this.getClass().getName() +")");
    }
    
    public final boolean equals(Object o) { return this == o || ((o instanceof JS) && jsequals((JS)o)); }
    
    // FIXME: Remove this, eventually
    public final String toString() { throw new Error("you shouldn't use toString() on JS objects"); }
    
    // These are only used internally by org.ibex.js
    Trap getTrap(JS key) { return null; }
    void putTrap(JS key, Trap value) throws JSExn { throw new JSExn("traps cannot be placed on this object (class=" + this.getClass().getName() +")"); }
    String coerceToString() throws JSExn { throw new JSExn("can't coerce to a string (class=" + getClass().getName() +")"); }
    boolean jsequals(JS o) { return this == o; }
    
    public static class O extends JS {
        private Hash entries;
        
        public Enumeration keys() throws JSExn { return entries == null ? emptyEnumeration : entries.keys(); }
        public JS get(JS key) throws JSExn { return entries == null ? null : (JS)entries.get(key, null); }
        public void put(JS key, JS val) throws JSExn { (entries==null?entries=new Hash():entries).put(key,null,val); }        

        /** retrieve a trap from the entries hash */
        final Trap getTrap(JS key) {
            return entries == null ? null : (Trap)entries.get(key, Trap.class);
        }
        
        /** retrieve a trap from the entries hash */
        final void putTrap(JS key, Trap value) {
            if (entries == null) entries = new Hash();
            entries.put(key, Trap.class, value);
        }    
    }
    
    public static class BT extends O {
        private BalancedTree bt;
        private final BalancedTree bt() { if(bt != null) return bt; return bt = new BalancedTree(); }
        public final void insertNode(int index, Object o) { bt().insertNode(index,o); }
        public final void clear() { bt().clear(); }
        public final Object getNode(int i) { return bt().getNode(i); }
        public final int treeSize() { return bt().treeSize(); }
        public final Object deleteNode(int i) { return bt().deleteNode(i); }
        public final void replaceNode(int index, Object o) { bt().replaceNode(index,o); }
        public final int indexNode(Object o) { return bt().indexNode(o); }
    }
    
    // FEATURE: JS.StringKeys
    /* public static StringKeys extends JS {
        public JS get(JS key) { return JS.isString(key) ? get(JS.toString(key) : null; }
        ...
    */
    
    JS _unclone() { return this; }
    
    public interface Cloneable { }
    
    public static class Clone extends JS.O implements Cloneable {
        protected final JS clonee;
        JS _unclone() { return clonee.unclone(); }
        public JS getClonee() { return clonee; }
        public Clone(JS clonee) throws JSExn {
            if(!(clonee instanceof Cloneable)) throw new JSExn("" + getClass().getName() + " isn't cloneable");
            this.clonee = clonee;
        }
        public boolean jsequals(JS o) { return unclone().jsequals(o.unclone()); }
        public Enumeration keys() throws JSExn { return clonee.keys(); }
        public JS get(JS key) throws JSExn { return clonee.get(key); }
        public void put(JS key, JS val) throws JSExn { clonee.put(key, val); }
        public JS callMethod(JS method, JS a0, JS a1, JS a2, JS[] rest, int nargs) throws JSExn {
            return clonee.callMethod(method, a0, a1, a2, rest, nargs);
        }
        public JS call(JS a0, JS a1, JS a2, JS[] rest, int nargs) throws JSExn {
            return clonee.call(a0, a1, a2, rest, nargs);
        }
    }
    
    // Static Interpreter Control Methods ///////////////////////////////////////////////////////////////

    /** log a message with the current JavaScript sourceName/line */
    public static void log(Object message) { info(message); }
    public static void debug(Object message) { Log.debug(Interpreter.getSourceName() + ":" + Interpreter.getLine(), message); }
    public static void info(Object message) { Log.info(Interpreter.getSourceName() + ":" + Interpreter.getLine(), message); }
    public static void warn(Object message) { Log.warn(Interpreter.getSourceName() + ":" + Interpreter.getLine(), message); }
    public static void error(Object message) { Log.error(Interpreter.getSourceName() + ":" + Interpreter.getLine(), message); }

    public static class NotPauseableException extends Exception { NotPauseableException() { } }

    /** returns a callback which will restart the context; expects a value to be pushed onto the stack when unpaused */
    public static UnpauseCallback pause() throws NotPauseableException {
        Interpreter i = Interpreter.current();
        if (i.pausecount == -1) throw new NotPauseableException();
        boolean get;
        switch(i.f.op[i.pc]) {
            case Tokens.RETURN: case ByteCodes.PUT: get = false; break;
            case ByteCodes.GET: get = true; break;
            default: throw new Error("should never happen");
        }
        i.pausecount++;
        return new JS.UnpauseCallback(i,get);
    }

    public static class UnpauseCallback implements Task {
        private Interpreter i;
        private boolean get;
        UnpauseCallback(Interpreter i, boolean get) { this.i = i; this.get = get; }
        public void perform() throws JSExn { unpause(); }
        public JS unpause() throws JSExn { return unpause((JS)null); }
        public JS unpause(JS o) throws JSExn {
            if (o == JS.METHOD) throw new JSExn("can't return a method to a paused context");
            if(get) i.stack.push(o);
            return i.resume();
        }
        public JS unpause(JSExn e) throws JSExn {
            i.catchException(e);
            return i.resume();
        }
    }



    // Static Helper Methods ///////////////////////////////////////////////////////////////////////////////////

    /** coerce an object to a Boolean */
    public static boolean toBoolean(JS o) {
        if(o == null) return false;
        if(o instanceof JSNumber) return ((JSNumber)o).toBoolean();
        if(o instanceof JSString) return ((JSString)o).length() != 0;
        return true;
    }
    //#repeat long/int/double/float toLong/toInt/toDouble/toFloat Long/Integer/Double/Float parseLong/parseInt/parseDouble/parseFloat
    /** coerce an object to a Number */
    public static long toLong(JS o) throws JSExn {
        if(o == null) return 0;
        if(o instanceof JSNumber) return ((JSNumber)o).toLong();
        if(o instanceof JSString) return Long.parseLong(o.coerceToString());
        throw new JSExn("can't coerce a " + o.getClass().getName() + " to a number");
    }
    //#end
    
    public static String toString(JS o) throws JSExn {
        if(o == null) return "null";
        return o.coerceToString();
    }
    
    public static String debugToString(JS o) {
        try { return toString(o); }
        catch(JSExn e) { return "[class=" + o.getClass().getName() + "]"; }
    }
    
    public static boolean isInt(JS o) {
        if(o == null) return true;
        if(o instanceof JSNumber.I) return true;
        if(o instanceof JSNumber.B) return false;
        if(o instanceof JSNumber) {
            JSNumber n = (JSNumber) o;
            return n.toInt() == n.toDouble();
        }
        if(o instanceof JSString) {
            String s = ((JSString)o).s;
            for(int i=0;i<s.length();i++)
                if(s.charAt(i) < '0' || s.charAt(i) > '9') return false;
            return true;
        }
        return false;
    }
    
    public static boolean isString(JS o) {
        if(o instanceof JSString) return true;
        return false;
    }

    // Instance Methods ////////////////////////////////////////////////////////////////////
 
    public final static JS NaN = new JSNumber.D(Double.NaN);
    public final static JS ZERO = new JSNumber.I(0);
    public final static JS ONE = new JSNumber.I(1);
        
    public static final JS T = new JSNumber.B(true);
    public static final JS F = new JSNumber.B(false);

    public static final JS B(boolean b) { return b ? T : F; }
    public static final JS B(int i) { return i==0 ? F : T; }
    
    private static final int CACHE_SIZE = 65536 / 4; // must be a power of two
    private static final JSString[] stringCache = new JSString[CACHE_SIZE];
    public static final JS S(String s)  {
        if(s == null) return null;
        int slot = s.hashCode()&(CACHE_SIZE-1);
        JSString ret = stringCache[slot];
        if(ret == null || !ret.s.equals(s)) stringCache[slot] = ret = new JSString(s);
        return ret;
    }

    public static final JS N(double d) { return new JSNumber.D(d); }
    public static final JS N(long l) { return new JSNumber.L(l); }
    
    public static final JS N(Number n) {
        if(n instanceof Integer) return N(n.intValue());
        if(n instanceof Long) return N(n.longValue());
        return N(n.doubleValue());
    }

    private static final JSNumber.I[] smallIntCache = new JSNumber.I[CACHE_SIZE];
    private static final JSNumber.I[] largeIntCache = new JSNumber.I[CACHE_SIZE];
    public static final JS N(int i) {
        JSNumber.I ret = null;
        int idx = i + smallIntCache.length / 2;
        if (idx < CACHE_SIZE && idx > 0) {
            ret = smallIntCache[idx];
            if (ret != null) return ret;
        }
        else ret = largeIntCache[Math.abs(idx % CACHE_SIZE)];
        if (ret == null || ret.i != i) {
            ret = new JSNumber.I(i);
            if (idx < smallIntCache.length && idx > 0) smallIntCache[idx] = ret;
            else largeIntCache[Math.abs(idx % CACHE_SIZE)] = ret;
        }
        return ret;
    }
    
    private static Enumeration emptyEnumeration = new Enumeration() {
            public boolean hasMoreElements() { return false; }
            public Object nextElement() { throw new NoSuchElementException(); }
        };
    
    public static JS fromReader(String sourceName, int firstLine, Reader sourceCode) throws IOException {
        return JSFunction._fromReader(sourceName, firstLine, sourceCode);
    }

    // HACK: caller can't know if the argument is a JSFunction or not...
    public static JS cloneWithNewParentScope(JS j, JSScope s) {
        return ((JSFunction)j)._cloneWithNewParentScope(s);
    }


    // Trap support //////////////////////////////////////////////////////////////////////////////

    /** performs a put, triggering traps if present; traps are run in an unpauseable interpreter */
    public void putAndTriggerTraps(JS key, JS value) throws JSExn {
        Trap t = getTrap(key);
        if (t != null) t.invoke(value);
        else put(key, value);
    }

    /** performs a get, triggering traps if present; traps are run in an unpauseable interpreter */
    public JS getAndTriggerTraps(JS key) throws JSExn {
        Trap t = getTrap(key);
        if (t != null) return t.invoke();
        else return get(key);
    }

    /** adds a trap, avoiding duplicates */
    final void addTrap(JS key, JSFunction f) throws JSExn {
        if (f.numFormalArgs > 1) throw new JSExn("traps must take either one argument (write) or no arguments (read)");
        boolean isRead = f.numFormalArgs == 0;
        for(Trap t = getTrap(key); t != null; t = t.next) if (t.f == f) return;
        putTrap(key, new Trap(this, key, f, (Trap)getTrap(key)));
    }

    /** deletes a trap, if present */
    final void delTrap(JS key, JSFunction f) throws JSExn {
        Trap t = (Trap)getTrap(key);
        if (t == null) return;
        if (t.f == f) { putTrap(t.name, t.next); return; }
        for(; t.next != null; t = t.next) if (t.next.f == f) { t.next = t.next.next; return; }
    }


} 
