// Copyright 2004 Adam Megacz, see the COPYING file for licensing [GPL]
package org.ibex.js;

import java.io.*;
import java.util.zip.*;
import org.ibex.util.*;
import org.ibex.plat.*;
import org.ibex.net.*;

/**
 *   Essentiall an InputStream "factory".  You can repeatedly ask a
 *   Stream for an InputStream, and each InputStream you get back will
 *   be totally independent of the others (ie separate stream position
 *   and state) although they draw from the same data source.
 */
public abstract class Stream extends JS implements JS.Cloneable {

    // Public Interface //////////////////////////////////////////////////////////////////////////////

    // FIXME: This should be in JS, "everything has a stream"
    public static InputStream getInputStream(JS js) throws IOException { return ((Stream)js.unclone()).getInputStream();}
    public static class NotCacheableException extends Exception { }

    private Cache getCache = new Cache(100);
    // FEATURE: Mandate that Streams use only String keys?
    protected JS _get(JS key) throws JSExn { return null; }
    public final JS get(JS key) throws JSExn {
        JS ret = (JS) getCache.get(key);
        if (ret == null) getCache.put(key, ret = _get(key));
        return ret;
    }

    // Private Interface //////////////////////////////////////////////////////////////////////////////

    public abstract InputStream getInputStream() throws IOException;
    protected String getCacheKey() throws NotCacheableException { throw new NotCacheableException(); }

    /** HTTP or HTTPS resource */
    // FEATURE: Only instansiate only ibex.net.HTTP, share with all substreams
    public static class HTTP extends Stream {
        private String url;
        public String coerceToString() { return "Stream.HTTP:" + url; }
        public HTTP(String url) { while (url.endsWith("/")) url = url.substring(0, url.length() - 1); this.url = url; }
        public JS _get(JS key) throws JSExn { return new HTTP(url + "/" + JS.toString(key)); }
        public String getCacheKey(Vec path) throws NotCacheableException { return url; }
        public InputStream getInputStream() throws IOException { return new org.ibex.net.HTTP(url).GET(); }
    }

    /** byte arrays */
    public static class ByteArray extends Stream {
        private byte[] bytes;
        private String cacheKey;
        public ByteArray(byte[] bytes, String cacheKey) { this.bytes = bytes; this.cacheKey = cacheKey; }
        public String getCacheKey() throws NotCacheableException {
            if (cacheKey == null) throw new NotCacheableException(); return cacheKey; }
        public InputStream getInputStream() throws IOException { return new ByteArrayInputStream(bytes); }
    }

    /** a file */
    public static class File extends Stream {
        private String path;
        public File(String path) { this.path = path; }
        public String coerceToString() { return "file:" + path; }
        public String getCacheKey() throws NotCacheableException { throw new NotCacheableException(); /* already on disk */ }
        public InputStream getInputStream() throws IOException { return new FileInputStream(path); }
        public JS _get(JS key) throws JSExn { return new File(path + java.io.File.separatorChar + JS.toString(key)); }
    }

    /** "unwrap" a Zip archive */
    public static class Zip extends Stream {
        private Stream parent;
        private String path;
        public Zip(Stream parent) { this(parent, null); }
        public Zip(Stream parent, String path) {
            while(path != null && path.startsWith("/")) path = path.substring(1);
            this.parent = parent;
            this.path = path;
        }
        public String getCacheKey() throws NotCacheableException { return parent.getCacheKey() + "!zip:"; }
        public JS _get(JS key) throws JSExn { return new Zip(parent, path==null?JS.toString(key):path+'/'+JS.toString(key)); }
        public InputStream getInputStream() throws IOException {
            InputStream pis = parent.getInputStream();
            ZipInputStream zis = new ZipInputStream(pis);
            ZipEntry ze = zis.getNextEntry();
            while(ze != null && !ze.getName().equals(path)) ze = zis.getNextEntry();
            if (ze == null) throw new IOException("requested file (" + path + ") not found in archive");
            return new KnownLength.KnownLengthInputStream(zis, (int)ze.getSize());
        }
    }

    /** "unwrap" a Cab archive */
    public static class Cab extends Stream {
        private Stream parent;
        private String path;
        public Cab(Stream parent) { this(parent, null); }
        public Cab(Stream parent, String path) { this.parent = parent; this.path = path; }
        public String getCacheKey() throws NotCacheableException { return parent.getCacheKey() + "!cab:"; }
        public JS _get(JS key) throws JSExn { return new Cab(parent, path==null?JS.toString(key):path+'/'+JS.toString(key)); }
        public InputStream getInputStream() throws IOException { return new MSPack(parent.getInputStream()).getInputStream(path); }
    }

    /** the Builtin resource */
    public static class Builtin extends Stream {
        public String getCacheKey() throws NotCacheableException { throw new NotCacheableException(); }
        public InputStream getInputStream() throws IOException { return Platform.getBuiltinInputStream(); }
    }

    /** shadow resource which replaces the graft */
    public static class ProgressWatcher extends Stream {
        final Stream watchee;
        JS callback;
        public ProgressWatcher(Stream watchee, JS callback) { this.watchee = watchee; this.callback = callback; }
        public String getCacheKey() throws NotCacheableException { return watchee.getCacheKey(); }
        public InputStream getInputStream() throws IOException {
            final InputStream is = watchee.getInputStream();
            return new FilterInputStream(is) {
                    int bytesDownloaded = 0;
                    public int read() throws IOException {
                        int ret = super.read();
                        if (ret != -1) bytesDownloaded++;
                        return ret;
                    }
                    public int read(byte[] b, int off, int len) throws IOException {
                        int ret = super.read(b, off, len);
                        if (ret != 1) bytesDownloaded += ret;
                        Scheduler.add(new Task() { public void perform() throws IOException, JSExn {
                            callback.call(N(bytesDownloaded),
                                          N(is instanceof KnownLength ? ((KnownLength)is).getLength() : 0), null, null, 2);
                        } });
                        return ret;
                    }
                };
        }
    }

    /** subclass from this if you want a CachedInputStream for each path */
    public static class CachedStream extends Stream {
        private Stream parent;
        private boolean disk = false;
        private String key;
        public String getCacheKey() throws NotCacheableException { return key; }
        CachedInputStream cis = null;
        public CachedStream(Stream p, String s, boolean d) throws NotCacheableException {
            this.parent = p; this.disk = d; this.key = p.getCacheKey();
        }
        public InputStream getInputStream() throws IOException {
            if (cis != null) return cis.getInputStream();
            if (!disk) {
                cis = new CachedInputStream(parent.getInputStream());
            } else {
                java.io.File f = org.ibex.core.LocalStorage.Cache.getCacheFileForKey(key);
                if (f.exists()) return new FileInputStream(f);
                cis = new CachedInputStream(parent.getInputStream(), f);
            }
            return cis.getInputStream();
        }
    }
}
